//! This module provides support for removing the extraneous break statements
//! generated by the incremental relooper.
use super::*;
use syn::{Expr, ExprBlock, ExprBreak, ExprIf, ExprLit, ExprMatch, ExprReturn, Lit};

pub struct IncCleanup {
    in_tail: Option<ImplicitReturnType>,
    brk_lbl: Label,
}

impl IncCleanup {
    pub fn new(in_tail: Option<ImplicitReturnType>, brk_lbl: Label) -> Self {
        IncCleanup { in_tail, brk_lbl }
    }

    /// The only way we can say for sure that we don't need a labelled block is if we remove
    /// the (unique) break to that label. We know that the label will be unique because relooper
    /// never duplicates blocks.
    ///
    /// Returns true if we manage to remove a tail expr.
    pub fn remove_tail_expr(&self, stmts: &mut Vec<Stmt>) -> bool {
        if let Some(mut stmt) = stmts.pop() {
            // If the very last stmt in our relooped output is a return/break, we can just
            // remove that statement. We additionally know that there is definitely no need
            // to label a block (if we were in that mode in the first place).
            if self.is_idempotent_tail_expr(&stmt) {
                return true;
            }

            let mut removed_tail_expr = false;

            if let Stmt::Expr(ref mut expr) = stmt {
                match expr {
                    Expr::If(ExprIf {
                        cond: _,
                        then_branch: ref mut body,
                        else_branch: ref mut sels,
                        ..
                    }) => {
                        removed_tail_expr =
                            removed_tail_expr || self.remove_tail_expr(&mut body.stmts);
                        if let Some((_token, els)) = sels {
                            if let Expr::Block(ExprBlock {
                                block: ref mut blk, ..
                            }) = **els
                            {
                                removed_tail_expr =
                                    removed_tail_expr || self.remove_tail_expr(&mut blk.stmts)
                            }
                        }
                    }

                    Expr::Match(ExprMatch {
                        arms: ref mut cases,
                        ..
                    }) => {
                        // Block label can be removed from any arm
                        for case in cases {
                            match *case.body {
                                Expr::Block(ExprBlock {
                                    block: ref mut blk, ..
                                }) => {
                                    removed_tail_expr =
                                        removed_tail_expr || self.remove_tail_expr(&mut blk.stmts)
                                }
                                _ => (),
                            }
                        }
                    }

                    _ => (),
                }
            }

            stmt = cleanup_if(stmt);

            // In all other cases, we give up and accept that we can't get rid of the last
            // stmt and that we might need a block label.
            stmts.push(stmt);
            removed_tail_expr
        } else {
            false
        }
    }

    fn is_idempotent_tail_expr(&self, stmt: &Stmt) -> bool {
        let tail_expr = if let Stmt::Semi(ref expr, _token) = *stmt {
            expr
        } else {
            return false;
        };
        match self.in_tail {
            Some(ImplicitReturnType::Main) => {
                if let Expr::Return(ExprReturn {
                    expr: Some(ref zero),
                    ..
                }) = *tail_expr
                {
                    if let Expr::Lit(ExprLit {
                        lit: Lit::Int(ref lit),
                        ..
                    }) = **zero
                    {
                        if let "0" = lit.base10_digits() {
                            return true;
                        }
                    }
                }
                false
            }

            Some(ImplicitReturnType::Void) => {
                if let Expr::Return(ExprReturn { expr: None, .. }) = *tail_expr {
                    return true;
                }
                false
            }

            _ => {
                if let Expr::Break(ExprBreak {
                    label: Some(ref blbl),
                    expr: None,
                    ..
                }) = *tail_expr
                {
                    if blbl.ident == mk().label(self.brk_lbl.pretty_print()).name.ident {
                        return true;
                    }
                }
                false
            }
        }
    }
}

/// Remove empty else clauses from if expressions that can arise from
/// removing idempotent statements.
fn cleanup_if(stmt: Stmt) -> Stmt {
    if let Stmt::Expr(ref expr) = &stmt {
        if let Expr::If(ExprIf {
            ref cond,
            then_branch: ref body,
            else_branch: ref els,
            ..
        }) = *expr
        {
            if let Some((_token, ref els)) = *els {
                if let Expr::Block(ExprBlock {
                    block: ref blk,
                    label: None,
                    ..
                }) = **els
                {
                    if blk.stmts.is_empty() {
                        return Stmt::Expr(Expr::If(ExprIf {
                            attrs: vec![],
                            if_token: Default::default(),
                            cond: cond.clone(),
                            then_branch: body.clone(),
                            else_branch: None,
                        }));
                    }
                }
            }
        }
    }
    stmt
}
